;; Generated package description from exwm.el  -*- no-byte-compile: t -*-
(define-package "exwm" "0.24.0.20200712.0" "Emacs X Window Manager" '((xelb "0.18")) :keywords '("unix") :authors '(("Chris Feng" . "chris.w.feng@gmail.com")) :maintainer '("Chris Feng" . "chris.w.feng@gmail.com") :url "https://github.com/ch11ng/exwm")
